#ifndef CPPBIND_CONNECTION_H
#define CPPBIND_CONNECTION_H

#include "os_defs.h"

#include <string>

#include <sql.h>
#include <sqlext.h>
#include <sqlucode.h>

#include "abs_types.h"

#ifndef WIN32
#define SQL_API
#endif

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus
   SQLRETURN SQL_API cpp_binding_connect(SQLHDBC conn,
                                         const wchar_t* host,
                                         const wchar_t* port,
                                         const wchar_t* nsp,
                                         const wchar_t* user,
                                         const wchar_t* pwd,
                                         bool* uni_srv);

   SQLRETURN SQL_API cpp_binding_alloc_messenger(SQLHDBC, void**);
   SQLRETURN SQL_API cpp_binding_free_messenger(void* pm);
   SQLRETURN SQL_API cpp_binding_get_header(void* pm, UINT4* msg_len, bool* last_piece);
   SQLRETURN SQL_API cpp_binding_put_header(void* pm, const char* msg, UINT4 msg_len,
                                            char* buf);
   SQLRETURN SQL_API cpp_binding_receive(void* pm, UINT4 len, char* buf);
   SQLRETURN SQL_API cpp_binding_send(void* pm, const char* buf, UINT4 len);
#ifdef __cplusplus
}
#endif // _#ifdef __cplusplus

//==============

class Conn_t;

typedef Handle<Conn_t> d_connection;

class Conn_t : public is_uni_srv_info {
private:
   bool connected;
   bool busy;

protected:

   void connect()
      { connected = true; }
   virtual void disconnect()
      { connected = false; }

   static bool parse_conn_str(const std::wstring& conn_str,
                              std::wstring* host,
                              std::wstring* port,
                              std::wstring* nsp);

   void destroy() // VC6 doesn't support virtual calls in destructors
      { if (connected) disconnect(); }

public:
   Conn_t()
      : connected(false), busy(false)
      {}
   virtual ~Conn_t()
      { destroy(); }

   // User interface

   bool is_connected() const
      { return connected; }
   bool is_busy() const
      { return busy; }

   // Database interface

   void lock()
      {
         if (!connected) throw Db_err(ERR_USER, "Connection is not active");
         if (busy) throw Db_err(ERR_USER, "Connection is busy");
         busy = true;
      }
   void unlock()
      { busy = false; }

   virtual int alloc_messenger(void** ppm)
      { throw Db_err(ERR_INTERNAL, "internal error in Conn_t::alloc_messenger()"); }
   virtual int free_messenger(void* pm)
      { throw Db_err(ERR_INTERNAL, "internal error in Conn_t::free_messenger()"); }

   static d_connection empty_conn_obj(Conn_t* conn,
                                      Db_err* err,
                                      int err_code,
                                      const char* err_msg)
      {
         if (err != 0)
         {
            err->set_code(err_code);
            err->set_msg(err_msg);
         }
         return d_connection(conn);
      }
};

//==============

class d_string;

class tcp_conn : public Conn_t {
   friend class Database;

private:
   SQLHENV henv;
   SQLHDBC hdbc;
   SQLHSTMT query_info;

   DLL_DECL tcp_conn(SQLHENV _henv, SQLHDBC _hdbc, bool uni_srv);

protected:
   DLL_DECL void disconnect();

public:
   tcp_conn()
      : henv(SQL_NULL_HENV), hdbc(SQL_NULL_HDBC), query_info(SQL_NULL_HSTMT)
      {}

   ~tcp_conn()
      { destroy(); }

   DLL_DECL static d_connection connect(const d_string& conn_str,
                                        const d_string& user,
                                        const d_string& pwd,
                                        Db_err* err = 0);

   int alloc_messenger(void** ppm);
   int free_messenger(void* pm);
};

//==============

#ifdef ActiveX_SUPPORT

DLL_DECL bool parse_com_conn_str(const wchar_t* com_conn_str,
                                 char_size_t size,
                                 d_string& conn_str,
                                 d_string& user,
                                 d_string& pwd);

DLL_DECL d_connection tcp_com_connect(const wchar_t* com_conn_str,
                                      char_size_t size,
                                      Db_err* err = 0);

#endif

//==============

#endif
