#ifndef CPPBIND_DATABASE_H
#define CPPBIND_DATABASE_H

#include "os_defs.h"

#include <map>
#include <set>
#include <string>
#include <typeinfo>

#include "class_def.h"
#include "cache_ant.h"
#include "connection.h"
#include "dlist.h"

class Obj_t;
class Database;

// for database::run_method
typedef enum {LIT_T = 0, OBJ_T=1, VOID_T=3} db_ret_type;

//==============

class Class_def_set {
private:
   std::map<db_name, Class_def*> defs;
   typedef std::map<db_name, Class_def*>::iterator It;

public:
   ~Class_def_set();

   DLL_DECL Class_def* open(Database* db, const_name_t type);
   void close(const_name_t type) {}
};

//==============

class Proxies_info {
private:
   class Key_t : public std::pair<int, std::string> {
   public:
      Key_t(int oref, const std::type_info& info)
         { first = oref; second = info.name(); }

      bool operator<(const Key_t& key) const
         { return (first < key.first) || ((first == key.first) && (second < key.second)); }
   };

   typedef std::pair<Obj_t*, int*> Val_t;
   std::map<Key_t, Val_t> info_map;

   std::map<int, int> ref_cnt_map;

public:
   DLL_DECL bool get_proxy_info(int oref, const std::type_info& info,
                                Obj_t** obj, int** num_refs);
   DLL_DECL void insert_proxy_info(int oref, const std::type_info& info,
                                   Obj_t* obj, int* num_refs);
   DLL_DECL void erase_proxy_info(int oref, const std::type_info& info);

   void inc_ref_cnt(int oref)
      {
         std::map<int, int>::iterator itor = ref_cnt_map.find(oref);
         if (itor != ref_cnt_map.end())
            ++(itor->second);
         else
            ref_cnt_map.insert(std::make_pair(oref, 1));
      }
   int dec_ref_cnt(int oref)
      {
         std::map<int, int>::iterator itor = ref_cnt_map.find(oref);
         if (itor == ref_cnt_map.end())
            throw Db_err(ERR_INTERNAL, "assertion failed in dec_ref_cnt()");
         int res = --(itor->second);
         if (res == 0) ref_cnt_map.erase(itor);
         return res;
      }

   bool is_swizzled(int oref) const
      { return ref_cnt_map.find(oref) != ref_cnt_map.end(); }
};

//==============

const int max_db_output_size = 1024 * 32;

template<typename T>
class db_istrbuf : public std::basic_streambuf<T> {
public:
   db_istrbuf(const T* buf, std::streamsize size)
      { set_buf(buf, size); }
   db_istrbuf()
      {}

   void set_buf(const T* buf, std::streamsize size)
      { T* b = (T*) buf; setg(b, b, b + size); }
};

class db_output {
private:
   bool unicode;

   char bin_buf[max_db_output_size];
   byte_size_t bin_size;

   db_istrbuf<char> n_buf;
   db_istrbuf<wchar_t> w_buf;

   std::istream n_in;
   std::wistream w_in;

public:
   db_output()
      : n_buf(bin_buf, max_db_output_size),
        w_buf((wchar_t*) bin_buf, max_db_output_size / sizeof(wchar_t)),
        n_in(&n_buf),
        w_in(&w_buf)
      {}

   bool is_unicode() const
      { return unicode; }
   const char* get_bin_buf() const
      { return bin_buf; }
   byte_size_t get_bin_size() const
      { return bin_size; }

   DLL_DECL std::istream& get_istream();
   DLL_DECL std::wistream& get_wistream();

   // t_io

   void get(t_istream& in)
      { in.get(&unicode, bin_buf, max_db_output_size, &bin_size); }
};

inline t_istream& operator>>(t_istream& in, db_output& val)
{ val.get(in); return in; }

//==============

const double client_ver = 2.08;

#define INIT_DB_MSG "B0"
#define DESTROY_DB_MSG "B1"
#define GET_PROPERTIES_MSG "B2"
#define GET_METHODS_MSG "B3"
#define GET_QUERIES_MSG "B4"
#define INC_OBJ_REF_CNT_MSG "B5"
#define DEC_OBJ_REF_CNT_MSG "B6"
#define RUN_FUNC_MSG "B7"
#define RUN_PROC_MSG "B8"
#define OPEN_OBJ_MSG "B9"
#define DEL_OBJ_MSG "BA"
#define TRANS_MSG "BB"
#define INIT_PRIM_LST "BC"
#define INIT_PRIM_ARR "BD"
#define INIT_OBJ_LST "BE"
#define INIT_OBJ_ARR "BF"
#define OPENID_OBJ_MSG "BG"
#define GET_PROP_MSG "BH"
#define SET_PROP_MSG "BI"
#define STATUS_MSG "BJ"

const int max_num_obj_args = 128;

const int arg_ptrs_buf_cap = max_num_obj_args * sizeof(D_type*);
const int refs_buf_cap = max_num_obj_args * sizeof(int);

const int res_buf_cap = 1024 * 32;

class Database  {
   friend class Obj_t;
private:
   d_connection conn;
   HDBC hdbc;

   HSTMT query_info_stmt;

   char refs_buf[refs_buf_cap];
   char arg_ptrs_buf[arg_ptrs_buf_cap];

   void* messenger;
   t_itcpstream in;
   t_otcpstream out;

   db_output term_output;
   db_output_callback default_output_callback;
   db_output_callback* term_output_callback;

   char res_buf[res_buf_cap];
   t_istrstream res;

   Db_err err;
   bin_string srv_errs;

   Cache* cache;
   Class_def_set class_defs;
   Proxies_info proxies_info;

   db_name nsp;

   enum trans_flag_t { START = 1, COMMIT = 2, ROLLBACK = 3, LEVEL = 4 };

   void free_resources()
      {
         delete cache;
         if (messenger) conn->free_messenger(messenger);
         conn->unlock();
      }

   void get_srv_info();
   // assumes that connection is already established
   // and cache is initialized;

   DLL_DECL void send_msg(const char* msg);

   void put_common()
      {
         if (cache == 0)
            return;

         cache->put(out);
      }
   void get_common();

   DLL_DECL void throw_err(const Db_err& err);
   // throw the error of the right type that is based on its id;
   // this way it's easier to retrieve the data specific to the error

   void get_srv_err(const bin_string& err_data, Db_err* err);
   // there may be more than one server error; this method picks one
   // of them

   void inc_ref_cnt(int oref)
      { proxies_info.inc_ref_cnt(oref); }

   void dec_ref_cnt(int oref)
      { if (proxies_info.dec_ref_cnt(oref) == 0) dec_ref_cnt_low(oref); }

   void dec_ref_cnt_low(int oref);

   void get_from_db(int oref, int ii, int jj, db_ret_type is_obj, const_name_t name);

   void put_in_db(int oref,
                  int ii,
                  int jj,
                  int mod_flag,
                  bool is_obj,
                  const_name_t name,
                  const D_type* val);

   DLL_DECL void transaction(trans_flag_t flag, Db_err* err, int* level = 0);

public:
   DLL_DECL Database(const d_connection& conn, bool use_cache = true);
   DLL_DECL ~Database();

   d_connection get_conn()
      { return conn; }

   HDBC get_hdbc()
      { return hdbc; }

   db_output& get_term_output()
      { return term_output; }
   void set_term_output_callback(db_output_callback* f)
      { term_output_callback = f; }
   void reset_term_output_callback()
      { term_output_callback = &default_output_callback; }

   char* get_refs_buf()
      { return refs_buf; }
   // returns a buffer for an int array (indices of arguments
   // passed by reference

   char* get_arg_ptrs_buf()
      { return arg_ptrs_buf; }
   // returns a buffer for a D_type* array (array of arguments to
   // run_method()

   Proxies_info* get_proxies_info()
      { return &proxies_info; }

   const db_name& get_nsp() const
      { return nsp; }
   bool unicode_srv() const
      { return conn->is_uni_srv(); }

   DLL_DECL t_istream& make_obj(const_name_t type_name,
                                const_str_t init_val = 0,
                                Db_err* err = 0);
   DLL_DECL t_istream& open_obj(const d_binary& oid,
                                int concurrency = -1,
                                int timeout = -1,
                                Db_err* err = 0);
   DLL_DECL t_istream& openid_obj(const_name_t name,
                                  const_str_t ident,
                                  int concurrency = -1,
                                  int timeout = -1,
                                  Db_err* err = 0);
   void del_obj(const d_binary& oid, int concurrency)
      {
         out << oid << concurrency;
         send_msg(DEL_OBJ_MSG);
         in >> err; if (err) throw_err(err);
      }

   DLL_DECL t_istream& run_method(int obj_ref,
                                  const_name_t cl_name,
                                  const_name_t mtd_name,
                                  D_type** args,
                                  int num_args,
                                  const int* refs,
                                  int num_refs,
                                  db_ret_type ret_t);

   DLL_DECL t_istream& get_property(int oref, int ii, int jj, db_ret_type is_obj,
                                    const_name_t name);
   DLL_DECL void set_property(int oref,
                              int ii,
                              int jj,
                              int mod_flag,
                              db_ret_type is_obj,
                              const_name_t name,
                              D_type* val);

   void tstart(Db_err* err = 0)
      { transaction(START, err); }
   void tcommit(Db_err* err = 0)
      { transaction(COMMIT, err); }
   void trollback(Db_err* err = 0)
      { transaction(ROLLBACK, err); }
   int tlevel(Db_err* err = 0)
      {
         int res;
         transaction(LEVEL, err, &res);
         return res;
      }

   void get_status_info(const d_status& status,
                        d_int& code,
                        d_string& msg,
                        const char* lang = "",
                        Db_err* err = 0);

   // interface for Class_def

   t_istream& get_properties(const_name_t class_name)
      {
         out << class_name;
         send_msg(GET_PROPERTIES_MSG);
         in >> err; if (err) throw_err(err);
         return in;
      }

   t_istream& get_queries(const_name_t class_name)
      {
         out << class_name;
         send_msg(GET_QUERIES_MSG);
         in >> err; if (err) throw_err(err);
         return in;
      }

   t_istream& get_methods(const_name_t class_name)
      {
         out << class_name;
         send_msg(GET_METHODS_MSG);
         in >> err; if (err) throw_err(err);
         return in;
      }

   t_istream& get_query_info(const wchar_t* class_name, const wchar_t* query_name);

   const Class_def* open_cl_def(const_name_t class_name)
      { return class_defs.open(this, class_name); }
   void close_cl_def(const_name_t class_name)
      { class_defs.close(class_name); }

   // interface for Collections

   template<typename C>
   void init_coln(int oref, C& coln)
      {
         out << oref;
         send_msg(C::init_msg());
         coln.init(in);
      }
};

#endif
