#ifndef CPPBIND_OBJ_TYPES_H
#define CPPBIND_OBJ_TYPES_H

#include "database.h"

//==============

class Dyn_obj;

class Obj_t {
private:
   Database* db;
   int ref;
   bool null_val;

   Obj_t& operator=(const Obj_t obj);

private:
   void init_ref(int r);
   void close_ref()
      { if (ref != 0) db->dec_ref_cnt(ref); ref = 0; null_val = true; }

protected:
   Obj_t()
      : db(0), ref(0), null_val(true)
      {}

   DLL_DECL void init_openref(Database* db, int ref);

public:
   DLL_DECL virtual ~Obj_t();
   // need to make Obj_t polymorphic in order to use it with d_ref

   int get_ref() const
      { return ref; }
   Database* get_db() const
      { return db; }

   // use _ before null method names to avoid wrong method resolution in abs_d_ref
   bool _is_null() const
      { return null_val; }
};

//==============

class Transient_t : public Obj_t {
protected:
   Transient_t() {}
};

//==============

class Persistent_t : public Obj_t {
protected:
   Persistent_t() {}

public:
   d_binary oid() const
      {
         t_istream& in =
            get_db()->run_method(get_ref(), empty_name, m_Oid, 0, 0, 0, 0, LIT_T);
         d_binary res(in);
         return res;
      }
   d_status save() const
      {
         t_istream& in =
            get_db()->run_method(get_ref(), empty_name, m_Save, 0, 0, 0, 0, LIT_T);
         d_status res(in);
         return res;
      }
};

//==============

class Embedded_t : public Obj_t {
protected:
   Embedded_t() {}
};

//==============

class abs_d_ref : public D_type {
// derived from D_type to allow passing d_ref<T>* as D_type* into
// Database::run_method()  (pass by reference for objects)

private:
   Obj_t* obj;
   int* num_refs;

   DLL_DECL static abs_d_ref null_d_ref;

   bool init_if_null()
      {
         if ((obj == 0) || obj->_is_null())
         {
            D_type::make_null();
            return true;
         }
         return false;
      }

   void release_obj()
      {
         if (!is_null())
            obj->get_db()->get_proxies_info()->erase_proxy_info(obj->get_ref(),
                                                                typeid(*obj));
         delete obj;
         delete num_refs;
      }

protected:
   abs_d_ref(Obj_t* p)
      : obj(p), num_refs(new int(1))
      {
         if (!init_if_null())
            obj->get_db()->get_proxies_info()->insert_proxy_info(obj->get_ref(),
                                                                 typeid(*obj),
                                                                 obj,
                                                                 num_refs);
      }
   abs_d_ref(Obj_t* p, int* ref_cnt)
      : obj(p), num_refs(ref_cnt)
      {
         ++(*num_refs);
         init_if_null();
      }

   DLL_DECL abs_d_ref& operator=(const abs_d_ref& r);
   void set_to_me(abs_d_ref& r) const
      { r = *this; }

   Obj_t* operator->() const
      {
         if (is_null())
            throw Db_err(ERR_NULL_VALUE, "cannot dereference a null value");
         return obj;
      }

   friend std::ostream& operator<<(std::ostream& out, const abs_d_ref& r)
      { return r.is_null() ? out << "NULL" : out << r->get_ref(); }

public:
   abs_d_ref(const abs_d_ref& r)
      : obj(r.obj), num_refs(r.num_refs)
      {
         ++(*num_refs);
         *static_cast<D_type*>(this) = r;
      }

   ~abs_d_ref()
      { if (--(*num_refs) == 0) release_obj(); }

   bool operator==(const abs_d_ref& r) const
      {
         return
            (is_null() && r.is_null()) ||
            (!is_null() && !r.is_null() &&
             (obj->get_ref() == r.obj->get_ref()) &&
             (obj->get_db() == r.obj->get_db()));
      }
   bool operator!=(const abs_d_ref& r) const
      { return !(*this == r); }

   // D_type interface

   void make_null()
      { *this = null_d_ref; }

   void get(t_istream& in, Database* db)
      { throw Db_err(ERR_INTERNAL, "assertion failed in abs_d_type"); }
   // needed for null_d_ref

   void put(t_ostream& out, Database* db = 0) const
      { is_null() ? out.put_null() : out.put(__int64(obj->get_ref())); }

   DLL_DECL static void get_oref_n_name(t_istream& in, int* oref, const_name_t* name);

   // interface for ODBC

   static SQLSMALLINT get_cpp_type()
      { return SQL_C_SLONG; }
};

//==============

template<class T>
class d_ref : public abs_d_ref {
public:
   T* operator->() const
      { return (T*) abs_d_ref::operator->(); } // don't care about the kind of inheritance
   T& operator*() const
      { return *operator->(); }

   d_ref() : abs_d_ref(0) {}
   d_ref(T* p) : abs_d_ref(p) {}
   d_ref(T* p, int* ref_cnt) : abs_d_ref(p, ref_cnt) {}

   const T* get_conv_ptr() const
      { return 0; }
   template<typename P>
   d_ref& operator=(const d_ref<P>& p)
      { const T* conv_ptr = p.get_conv_ptr(); abs_d_ref::operator=(p); return *this; }

   template<typename P>
   d_ref(const d_ref<P>& p)
      : abs_d_ref(p)
      { const T* conv_ptr = p.get_conv_ptr(); }

   template<typename P>
   void conv_to(d_ref<P>& res)
      {
         const P* conv_ptr = dynamic_cast<const P*>(abs_d_ref::operator->());
         if (conv_ptr != 0) set_to_me(res);
         else res.make_null();
      }

   // T() should be private so that the user cannot instantiate T directly;
   // d_ref(T* p) is public because otherwise it is impossible to use
   // T::make_new() etc.

   d_ref(const d_ref& r) : abs_d_ref(r) {}

   // D_type interface

   // t_io

   void get(t_istream& in, Database* db);
};

template<typename T>
void d_ref<T>::get(t_istream& in, Database* db)
{
   int oref; const_name_t tname;
   get_oref_n_name(in, &oref, &tname);

   if (oref == 0)
      make_null();
   else if (!is_null() && ((*this)->get_ref() == oref))
      return;
   else
      *this = T::openref(db, oref, tname);
}

#endif
