#ifndef CPPBIND_PRIM_TYPES_H
#define CPPBIND_PRIM_TYPES_H

#include "os_defs.h"

#include <sql.h>
#include <sqlext.h>
#include <sqlucode.h>

#include <string>
#include <ctime>

#include "abs_types.h"
#include "uni_conv.h"

class Database;

bool tm_to_date_int(const tm& d, int& res);

//==============

class d_binary : public D_type {
protected:
   bin_string val;

public:
   d_binary()
      {}
   d_binary(const char* cstr)
      : val(cstr)
      {}
   d_binary(const char* cstr, int sz)
      : val(cstr, sz)
      {}
   d_binary(const std::string& s)
      : val(s)
      {}
   d_binary(const bin_string& s)
      : val(s)
      {}

   operator std::string() const
      { return val; }

   bool operator==(const d_binary& t)
      { return val == t.val; }
   bool operator!=(const d_binary& t)
      { return val != t.val; }

   void append_bin(const char* buf, byte_size_t size)
      { make_not_null(); val.append(buf, size); }

   // D_type interface

   void assign(const char* buf, byte_size_t size)
      { make_not_null(); val.assign(buf, size); }
   void make_null()
      { assign("", 0); D_type::make_null(); }

   const char* get_buf() const
      { return val.get_buf(); }
   long get_size() const
      { return val.get_size(); }

   // d_query interface

   static SQLSMALLINT get_cpp_type()
      { return SQL_C_BINARY; }

   // t_io

   void get(t_istream& in, Database* db = 0)
      { make_not_null(); bool uni; in.get(&uni, &val); }

   void put(t_ostream& out, Database* db = 0) const
      {
         is_null() ?
            out.put_null() :
            out.put(multibyte, val.get_buf(), (byte_size_t) val.get_size());
      }

   d_binary(t_istream& in)
      { get(in); }

};

inline t_ostream& operator<<(t_ostream& out, const d_binary& t)
{ t.put(out); return out; }

//==============

class d_string : public d_binary {
private:
   uni_t type;

public:
   // buffers for conversions
   static char buf1[conv_buf_size];
   static char buf2[conv_buf_size];

public:
   uni_t get_type() const
      { return type; }

   d_string()
      { type = multibyte; }
   d_string(const char* cstr)
      : d_binary(cstr)
      { type = multibyte; }
   d_string(const char* cstr, int sz)
      : d_binary(cstr, sz)
      { type = multibyte; }
   d_string(const wchar_t* cstr)
      : d_binary((const char*) cstr, wcslen(cstr) * sizeof(wchar_t))
      { type = cpp_uni; }
   d_string(const wchar_t* cstr, int sz)
      : d_binary((const char*) cstr, sz * sizeof(wchar_t))
      { type = cpp_uni; }
   d_string(const std::string& s)
      : d_binary(s)
      { type = multibyte; }
   d_string(const std::wstring& s)
      : d_binary((const char*) s.data(), s.length() * sizeof(wchar_t))
      { type = cpp_uni; }

   bool is_unicode() const
      { return (type == cache_uni) || (type == cpp_uni); }

   DLL_DECL byte_size_t to_mb(char* buf, char_size_t cap) const;
   DLL_DECL char_size_t to_uni(wchar_t* buf, char_size_t cap) const;

   DLL_DECL void to_uni();
   DLL_DECL void to_mb();

   operator std::string() const
      {
         char_size_t size = to_mb(conv_buf, conv_buf_size);
         return std::string(conv_buf, size);
      }
   operator std::wstring() const
      {
         wchar_t* buf = (wchar_t*) conv_buf;
         char_size_t cap = conv_buf_size / sizeof(wchar_t);
         char_size_t size = to_uni(buf, cap);
         return std::wstring(buf, size);
      }

   DLL_DECL bool operator==(const d_string& val) const;
   bool operator!=(const d_string& val) const
      { return !(*this == val); }
   DLL_DECL bool operator<(const d_string& val) const;

   void assign(const char* buf, char_size_t size)
      { d_binary::assign(buf, size); type = multibyte; }
   void assign(const char* buf)
      { assign(buf, strlen(buf)); }
   void assign(const wchar_t* buf, char_size_t size)
      { d_binary::assign((const char*) buf, size * sizeof(wchar_t)); type = cpp_uni; }
   void assign(const wchar_t* buf)
      { assign(buf, wcslen(buf)); }

   // d_query interface

   SQLSMALLINT get_cpp_type() const
      {
#ifdef CACHE_UNI
         if ((type == cache_uni) || (type == cpp_uni)) return SQL_C_WCHAR;
#else
         if (type == cpp_uni) return SQL_C_WCHAR;
#endif
         if (type == multibyte) return SQL_C_CHAR;
         return SQL_C_BINARY;
      }

   // t_io

   void get(t_istream& in, Database* db = 0)
      {
         bool uni;
         make_not_null(); in.get(&uni, &val);
         type = uni ? cache_uni : multibyte;
      }

   DLL_DECL void put(t_ostream& out, Database* db = 0) const;

   d_string(t_istream& in)
      { get(in); }
};

DLL_DECL std::ostream& operator<<(std::ostream& out, const d_string& t);

inline t_ostream& operator<<(t_ostream& out, const d_string& t)
{ t.put(out); return out; }

inline t_istream& operator>>(t_istream& in, d_string& val)
{ val.get(in); return in; }

//==============

class d_wstring : public D_type {
private:
   std::wstring val;

   const char* get_buf() const
      { return (char*) val.c_str(); }
   long get_size() const
      { return val.length() * sizeof(wchar_t); }

public:
   DLL_DECL static d_wstring cmn;

   d_wstring() {}
   d_wstring(const wchar_t* cstr)
      : val(cstr)
      {}
   d_wstring(const wchar_t* cstr, int sz)
      : val(cstr, sz)
      {}

   operator const wchar_t*() const
      { return val.c_str(); }
   int length() const
      { return val.length(); }

   void assign(const wchar_t* buf, char_size_t size)
      { make_not_null(); val.assign(buf, size); }
   void assign(const wchar_t* buf)
      { make_not_null(); val.assign(buf); }

   void append(const wchar_t* buf, char_size_t size)
      { make_not_null(); val.append(buf, size); }

   // d_query interface

   static SQLSMALLINT get_cpp_type()
      { return SQL_C_WCHAR; }

   DLL_DECL static int cmp(const wchar_t* p, const wchar_t* q);

   // t_io

   DLL_DECL void get(t_istream& in, Database* db = 0);
   DLL_DECL void put(t_ostream& out, Database* db = 0) const;
};

inline bool operator<(const d_wstring& s1, const d_wstring& s2)
{ return d_wstring::cmp(s1, s2) < 0; }

inline bool operator==(const d_wstring& s1, const d_wstring& s2)
{ return d_wstring::cmp(s1, s2) == 0; }

typedef d_wstring db_name;

//==============

class d_int : public D_type {
private:
   int val;

public:
   d_int()
      { make_null(); }
   d_int(int v)
      : val(v) {}

   int value() const { return val; }
   operator int() const { return val; }
   d_int& operator=(long new_val)
      { make_not_null(); val = new_val; return *this; }

   d_int& operator++()
      { if (!is_null()) ++val; return *this; }

   // D_type interface

   void make_null()
      { val = 0; D_type::make_null(); }

   // d_query interface

   static SQLSMALLINT get_cpp_type()
      { return SQL_C_SLONG; }

   // t_io

   void get(t_istream& in, Database* db = 0)
      {
         make_not_null();
         __int64 v; bool null; in.get(&v, &null);
         if (null) make_null();
         else val = int(v);
      }

   void put(t_ostream& out, Database* db = 0) const
      { is_null() ? out.put_null() : out.put(__int64(val)); }

   d_int(t_istream& in)
      { get(in); }
};

inline std::ostream& operator<<(std::ostream& out, const d_int& t)
{ return (!t.is_null()) ? out << t.value() : out << "NULL"; }

inline t_ostream& operator<<(t_ostream& out, const d_int& t)
{ t.put(out); return out; }

//==============

class d_double : public D_type {
private:
   double val;

public:
   d_double()
      { make_null(); }
   d_double(double v)
      : val(v) {}

   double value() const { return val; }
   operator double() const { return val; }
   d_double& operator=(double new_val)
      { make_not_null(); val = new_val; return *this; }

   // D_type interface

   void make_null()
      { val = 0; D_type::make_null(); }

   // d_query interface

   static SQLSMALLINT get_cpp_type()
      { return SQL_C_DOUBLE; }

   // t_io

   void get(t_istream& in, Database* db = 0)
      {
         make_not_null();
         bool null; in.get(&val, &null);
         if (null) make_null();
      }

   void put(t_ostream& out, Database* db = 0) const
      { is_null() ? out.put_null() : out.put(val); }

   d_double(t_istream& in)
      { get(in); }
};

inline std::ostream& operator<<(std::ostream& out, const d_double& t)
{ return (!t.is_null()) ? out << t.value() : out << "NULL"; }

inline t_ostream& operator<<(t_ostream& out, const d_double& t)
{ t.put(out); return out; }

//==============

class d_bool : public D_type {
private:
   bool val;

public:
   d_bool()
      { make_null(); }
   d_bool(bool v)
      : val(v) {}

   bool value() const { return val; }
   operator bool() const { return val; }
   d_bool& operator=(bool new_val)
      { make_not_null(); val = new_val; return *this; }

   d_bool& operator=(int new_val)
      { make_not_null(); val = (new_val != 0); return *this; }

   // D_type interface

   void make_null()
      { val = false; D_type::make_null(); }

   // d_query interface

   static SQLSMALLINT get_cpp_type()
      { return d_int::get_cpp_type(); }

   // t_io

   void get(t_istream& in, Database* db = 0)
      {
         make_not_null();
         __int64 v; bool null; in.get(&v, &null);
         if (null) make_null();
         else val = (v != 0);
      }

   void put(t_ostream& out, Database* db = 0) const
      { is_null() ? out.put_null() : out.put(__int64(val)); }

   d_bool(t_istream& in)
      { get(in); }
};

inline std::ostream& operator<<(std::ostream& out, const d_bool& t)
{ return (!t.is_null()) ? out << t.value() : out << "NULL"; }

inline t_ostream& operator<<(t_ostream& out, const d_bool& t)
{ t.put(out); return out; }

//==============

class d_currency : public D_type {
private:
   double val;

public:
   d_currency()
      { make_null(); }
   d_currency(double v)
      : val(v) {}

   double value() const { return val; }
   operator double() const { return val; }
   d_currency& operator=(double new_val)
      { make_not_null(); val = new_val; return *this; }

   // D_type interface

   void make_null()
      { val = 0; D_type::make_null(); }

   // d_query interface

   static SQLSMALLINT get_cpp_type()
      { return d_double::get_cpp_type(); }

   // t_io

   void get(t_istream& in, Database* db = 0)
      {
         make_not_null();
         bool null; in.get(&val, &null);
         if (null) make_null();
      }

   void put(t_ostream& out, Database* db = 0) const
      { is_null() ? out.put_null() : out.put(val); }

   d_currency(t_istream& in)
      { get(in); }
};

inline std::ostream& operator<<(std::ostream& out, const d_currency& t)
{ return (!t.is_null()) ? out << t.value() : out << "NULL"; }

inline t_ostream& operator<<(t_ostream& out, const d_currency& t)
{ t.put(out); return out; }

//==============

class d_status : public d_binary {
   friend class com_variant_traits;

private:
   d_int code;
   d_string msg;

   DLL_DECL void init();

public:
   d_status()
      : d_binary()
      {}
   d_status(const char* cstr)
      : d_binary(cstr)
      { init(); }
   d_status(const char* cstr, int sz)
      : d_binary(cstr, sz)
      { init(); }
   d_status(const bin_string& s)
      : d_binary(s)
      { init(); }

   operator int() const
      { return code; }
   int get_code() const
      { return code; }
   const d_string& get_msg() const
      { return msg; }

   DLL_DECL void get_from_srv(Database* db, const char* lang = "", Db_err* err = 0);

   void throw_err() const
      { throw Db_err(code, msg); }

   // D_type interface

   // t_io

   void get(t_istream& in, Database* db = 0)
      { d_binary::get(in, db); init(); }

   d_status(t_istream& in)
      { get(in); init(); }

};

inline std::ostream& operator<<(std::ostream& out, const d_status& t)
{
   if (t.is_null()) return out << "NULL";
   return (!t) ?
      out << "no errors" :
      out << "status code: " << t.get_code() << ", status error: "
          << std::string(t.get_msg());
}

inline t_ostream& operator<<(t_ostream& out, const d_status& t)
{ t.put(out); return out; }

//==============

class timestamp_struct {
   friend class com_variant_traits;

protected:
   tm val;

protected:
   void reset()
      {
         val.tm_sec = val.tm_min = val.tm_hour =
            val.tm_mday = val.tm_mon = val.tm_year =
            val.tm_wday = val.tm_yday = val.tm_isdst = -1;
      }

public:
   timestamp_struct()
      { reset(); }

   void conv_to(TIME_STRUCT& res) const
      {
         res.hour = val.tm_hour;
         res.minute = val.tm_min;
         res.second = val.tm_sec;
      }
   void conv_to(DATE_STRUCT& res) const
      {
         res.year = val.tm_year + 1900;
         res.month =val.tm_mon + 1;
         res.day = val.tm_mday;
      }
   void conv_to(TIMESTAMP_STRUCT& res) const
      {
         res.hour = val.tm_hour;
         res.minute = val.tm_min;
         res.second = val.tm_sec;
         res.year = val.tm_year + 1900;
         res.month = val.tm_mon + 1;
         res.day = val.tm_mday;
         res.fraction = 0;
      }
};

class d_time : public D_type, public timestamp_struct {
private:
   void init(const tm& t)
      { make_not_null(); val = t; }
   void init(const TIME_STRUCT& t)
      {
         make_not_null();
         val.tm_hour = t.hour; val.tm_min = t.minute; val.tm_sec = t.second;
      }
   DLL_DECL void init_low(int int_val);

public:
#ifdef ActiveX_SUPPORT
   DLL_DECL void odbc_init(const wchar_t* buf);
#endif

   d_time()
      { make_null(); }
   d_time(const tm& t)
      { init(t); }
   d_time(const TIME_STRUCT& t)
      { init(t); }
   d_time(int h, int m, int s)
      { val.tm_hour = h; val.tm_min = m; val.tm_sec = s; }

   d_time& operator=(const tm& t)
      { init(t); return *this; }
   d_time& operator=(const TIME_STRUCT& t)
      { init(t); return *this; }

   const tm& value() const { return val; }
   operator tm() const { return val; }

   // D_type interface

   void make_null()
      { reset(); D_type::make_null(); }

   // d_query interface

   static SQLSMALLINT get_cpp_type()
      { return SQL_C_TIME; }

   // t_io

   void get(t_istream& in, Database* db = 0)
      {
         make_not_null();
         __int64 v; bool null; in.get(&v, &null);
         if (null) make_null();
         else init_low(int(v));
      }

  void put(t_ostream& out, Database* db = 0) const
      {
         if (is_null()) out.put_null();
         else out.put(__int64(val.tm_hour * 3600 + val.tm_min * 60 + val.tm_sec));
      }

   d_time(t_istream& in)
      { get(in); }
};

DLL_DECL std::ostream& operator<<(std::ostream& out, const d_time&);

inline t_ostream& operator<<(t_ostream& out, const d_time& t)
{ t.put(out); return out; }

//==============

class d_date : public D_type, public timestamp_struct {
private:
   void init(const tm& d)
      { make_not_null(); val = d; }
   void init(const DATE_STRUCT& d)
      {
         make_not_null();
         val.tm_year = d.year - 1900; val.tm_mon = d.month - 1; val.tm_mday = d.day;
      }

   DLL_DECL void init_low(int int_val);

public:
#ifdef ActiveX_SUPPORT
   DLL_DECL void odbc_init(const wchar_t* buf);
#endif

   d_date()
      { make_null(); }
   d_date(const tm& d)
      { init(d); }
   d_date(const DATE_STRUCT& d)
      { init(d); }
   d_date(int y, int m, int d)
      { val.tm_year = y - 1900; val.tm_mon = m - 1; val.tm_mday = d; }

   d_date& operator=(const tm& d)
      { init(d); return *this; }
   d_date& operator=(const DATE_STRUCT& d)
      { init(d); return *this; }

   const tm& value() const { return val; }
   operator tm() const { return val; }

   // D_type interface

   void make_null()
      { reset(); D_type::make_null(); }

   // d_query interface

   static SQLSMALLINT get_cpp_type()
      { return SQL_C_DATE; }

   // t_io

   void get(t_istream& in, Database* db = 0)
      {
         make_not_null();
         __int64 v; bool null; in.get(&v, &null);
         if (null) make_null();
         else init_low(int(v));
      }

   DLL_DECL void put(t_ostream& out, Database* db = 0) const;

   d_date(t_istream& in)
      { get(in); }
};

DLL_DECL std::ostream& operator<<(std::ostream& out, const d_date& t);

inline t_ostream& operator<<(t_ostream& out, const d_date& t)
{ t.put(out); return out; }

//==============

class d_timestamp : public D_type, public timestamp_struct {
private:
   void init(const tm& ts)
      { make_not_null(); val = ts; }
   void init(const TIMESTAMP_STRUCT& ts)
      {
         make_not_null();
         val.tm_year = ts.year - 1900; val.tm_mon = ts.month - 1; val.tm_mday = ts.day;
         val.tm_hour = ts.hour; val.tm_min = ts.minute; val.tm_sec = ts.second;
      }

   DLL_DECL void init_low(const char* buf);
   DLL_DECL void init_low(const wchar_t* buf);

public:
#ifdef ActiveX_SUPPORT
   void odbc_init(const wchar_t* buf)
      { init_low(buf); }
#endif

   d_timestamp()
      { make_null(); }
   d_timestamp(const TIMESTAMP_STRUCT& ts)
      { init(ts); }
   d_timestamp(const tm& ts)
      { init(ts); }
   d_timestamp(const char* buf)
      { init_low(buf); }
   d_timestamp(const wchar_t* buf)
      { init_low(buf); }

   d_timestamp& operator=(const tm& ts)
      { init(ts); return *this; }
   d_timestamp& operator=(const TIMESTAMP_STRUCT& ts)
      { init(ts); return *this; }

   const tm& value() const { return val; }
   operator tm() const { return val; }

   // D_type interface

   void make_null()
      { reset(); D_type::make_null(); }

   // d_query interface

   static SQLSMALLINT get_cpp_type()
      { return SQL_C_TIMESTAMP; }

   // t_io

   DLL_DECL void get(t_istream& in, Database* db = 0);
   DLL_DECL void put(t_ostream& out, Database* db = 0) const;
   d_timestamp(t_istream& in)
      { get(in); }
};

DLL_DECL std::ostream& operator<<(std::ostream& out, const d_timestamp& t);

inline t_ostream& operator<<(t_ostream& out, const d_timestamp& t)
{ t.put(out); return out; }

//==============

typedef d_wstring d_id;
typedef d_binary d_oid;
typedef d_double d_numeric;

#endif
