#ifndef CPPBIND_QUERY_H
#define CPPBIND_QUERY_H

#include "os_defs.h"

#include <sql.h>
#include <sqlext.h>
#include <sqlucode.h>

#include "odbc_common.h"
#include "obj_types.h"

// the first parameter in many template functions here allows to use the
// template in developer studio; otherwise it fails to instantiate the
// template

// all __* functions are needed for VC6 -- it has a bug in instantiating
// functions defined in a different files

//==============

const int max_col_name_size = 256;
const int max_short_par_buf_len = 100;

struct Res_set_col_info {
   SQLWCHAR name[max_col_name_size];
   SQLSMALLINT name_len;
   SQLSMALLINT sql_type;
   SQLUINTEGER col_size;
   SQLINTEGER buf_size;
   SQLSMALLINT dec_digits;
   SQLSMALLINT nullable;
};

//==============

template<typename To, typename From>
inline void set_val(To* buf, const From& val)
{ *buf = val; }

template<>
void set_val(TIME_STRUCT* buf, const d_time& val);
template<>
void set_val(DATE_STRUCT* buf, const d_date& val);
template<>
void set_val(TIMESTAMP_STRUCT* buf, const d_timestamp& val);

typedef enum { NEED_TO_REBIND, LEAVE_AS_IS } par_status;

class Res_set_par {
public:
   SQLSMALLINT sql_type;
   SQLSMALLINT cpp_type;
   SQLUINTEGER col_size;
   SQLSMALLINT dec_digits;
   SQLSMALLINT nullable;

   bool unbound;

private:
   SQLINTEGER size;

   char* dyn_buf;
   int dyn_cap;

   union {
      long int_val;
      double double_val;
      TIME_STRUCT time_val;
      DATE_STRUCT date_val;
      TIMESTAMP_STRUCT timestamp_val;
      char st_buf[max_short_par_buf_len];
   };

   bool is_var_buf_t() const
      {
         return
            (cpp_type == SQL_C_CHAR) ||
            (cpp_type == SQL_C_WCHAR) ||
            (cpp_type == SQL_C_BINARY);
      }

   bool is_st_buf_in_use() const
      { return dyn_buf == 0; }

   par_status reset_dyn_buf(int val_size);

   template<typename To, typename From>
   par_status set_fixed_size_par(To* buf, const From& val);

   template<typename D_TYPE>
   par_status set_var_size_par(const D_TYPE& val);
   par_status set_var_size_par_low(par_status cur_status,
                                   const char* val_buf, byte_size_t val_size);

public:
   Res_set_par()
      : unbound(true), size(sizeof(int)),
        cpp_type(SQL_C_SLONG),
        dyn_buf(0), dyn_cap(0)
      {}
   ~Res_set_par()
      { delete[] dyn_buf; }

   const void* get_ptr() const
      {
         if (is_var_buf_t())
            return is_st_buf_in_use() ? st_buf : dyn_buf;
         else
            return &int_val;
      }

   const SQLINTEGER* get_psize() const
      { return &size; }
   int get_cap() const
      { return is_st_buf_in_use() ? max_short_par_buf_len : dyn_cap; }

   par_status set(const char* buf, char_size_t size)
      {
         par_status res = (cpp_type == SQL_C_CHAR) ? LEAVE_AS_IS : NEED_TO_REBIND;
         cpp_type = SQL_C_CHAR;
         return set_var_size_par_low(res, buf, (byte_size_t) size);
      }
   par_status set(const wchar_t* buf, char_size_t size)
      {
         par_status res = (cpp_type == SQL_C_WCHAR) ? LEAVE_AS_IS : NEED_TO_REBIND;
         cpp_type = SQL_C_WCHAR;
         return set_var_size_par_low(res, (char*) buf, size * sizeof(wchar_t));
      }

   par_status set()
      { size = SQL_NULL_DATA; return LEAVE_AS_IS; }

   par_status set(const d_int& val)
      { return set_fixed_size_par(&int_val, val); }
   par_status set(const d_double& val)
      { return set_fixed_size_par(&double_val, val); }
   par_status set(const d_bool& val)
      { return set_fixed_size_par(&int_val, val); }
   par_status set(const d_currency& val)
      { return set_fixed_size_par(&double_val, val); }
   par_status set(const d_time& val)
      { return set_fixed_size_par(&time_val, val); }
   par_status set(const d_date& val)
      { return set_fixed_size_par(&date_val, val); }
   par_status set(const d_timestamp& val)
      { return set_fixed_size_par(&timestamp_val, val); }
   par_status set(const d_string& val)
      { return set_var_size_par(val); }
   par_status set(const d_binary& val)
      { return set_var_size_par(val); }
};

template<typename To, typename From>
par_status Res_set_par::set_fixed_size_par(To* buf, const From& val)
{
   if (val.is_null())
   {
      size = SQL_NULL_DATA;
      return LEAVE_AS_IS;
   }
   else
   {
      par_status res = (cpp_type == val.get_cpp_type()) ? LEAVE_AS_IS : NEED_TO_REBIND;

      cpp_type = val.get_cpp_type();
      size = sizeof(To);
      set_val(buf, val);

      return res;
   }
}

template<typename D_TYPE>
par_status Res_set_par::set_var_size_par(const D_TYPE& val)
{
   if (val.is_null())
   {
      size = SQL_NULL_DATA;
      return LEAVE_AS_IS;
   }
   else
   {
      par_status res = (cpp_type == val.get_cpp_type()) ? LEAVE_AS_IS : NEED_TO_REBIND;

      cpp_type = val.get_cpp_type();

      int val_size = val.get_size();
      const char* val_buf = val.get_buf();
      return set_var_size_par_low(res, val_buf, val_size);
   }
}

//==============

typedef enum { CONV_TO_MB, CONV_TO_UNI, NO_CONV } str_conv_t;

class abs_d_query {
protected:
   SQLHSTMT hstmt;
   Database* db;

private:
   std::vector<Res_set_col_info> columns;
   std::vector<Res_set_par> pars;

   template<typename D_TYPE>
   void _set_par(int idx, const D_TYPE& val)
      {
         --idx;
         if ((pars[idx].set(val) == NEED_TO_REBIND) || pars[idx].unbound)
            set_par_low(idx);
      }
   template<typename STR_T>
   void _set_par(int idx, const STR_T* buf, char_size_t size)
      {
         --idx;
         if ((pars[idx].set(buf, size) == NEED_TO_REBIND)
             || pars[idx].unbound)
            set_par_low(idx);
      }

   void set_par_low(int idx)
      {
         SQLRETURN rc;
         rc = SQLBindParameter(hstmt,
                               idx + 1,
                               SQL_PARAM_INPUT,
                               pars[idx].cpp_type,
                               pars[idx].sql_type,
                               pars[idx].col_size,
                               pars[idx].dec_digits,
                               const_cast<void*>(pars[idx].get_ptr()),
                               pars[idx].get_cap(),
                               const_cast<SQLINTEGER*>(pars[idx].get_psize()));
         if (rc != SQL_SUCCESS) throw_err(SQL_HANDLE_STMT, hstmt);
         pars[idx].unbound = false;
      }

   void __set_par(int idx, const char* buf, char_size_t size)
      { _set_par(idx, buf, size); }
   void __set_par(int idx, const wchar_t* buf, char_size_t size)
      { _set_par(idx, buf, size); }

   void __set_par(int idx, const d_int& val) { _set_par(idx, val); }
   void __set_par(int idx, const d_double& val) { _set_par(idx, val); }
   void __set_par(int idx, const d_bool& val) { _set_par(idx, val); }
   void __set_par(int idx, const d_currency& val) { _set_par(idx, val); }
   void __set_par(int idx, const d_time& val) { _set_par(idx, val); }
   void __set_par(int idx, const d_date& val) { _set_par(idx, val); }
   void __set_par(int idx, const d_timestamp& val) { _set_par(idx, val); }
   void __set_par(int idx, const d_binary& val) { _set_par(idx, val); }
   void __set_par(int idx, const d_string& val) { _set_par(idx, val); }

protected:
   abs_d_query()
      : hstmt(SQL_NULL_HSTMT), db(0)
      {}

   DLL_DECL void set_conn(Database* db);
   DLL_DECL void release_conn();

   void throw_err(SQLSMALLINT err_src, SQLHANDLE handle);

public:
   abs_d_query(Database* db)
      { set_conn(db); }
   ~abs_d_query()
      { release_conn(); }

   DLL_DECL void prepare(const wchar_t* sql_query, int len);
   DLL_DECL void prepare(const wchar_t* cl_name, const wchar_t* proc_name);

   void prepare(d_string& sql_name, int num_pars);

   DLL_DECL void set_par(int idx, const char* buf, char_size_t size);
   DLL_DECL void set_par(int idx, const wchar_t* buf, char_size_t size);
   DLL_DECL void set_par(int idx, int val);
   DLL_DECL void set_par(int idx, double val);

   DLL_DECL void set_par(int idx, const d_int& val);
   DLL_DECL void set_par(int idx, const d_double& val);
   DLL_DECL void set_par(int idx, const d_bool& val);
   DLL_DECL void set_par(int idx, const d_currency& val);
   DLL_DECL void set_par(int idx, const d_time& val);
   DLL_DECL void set_par(int idx, const d_date& val);
   DLL_DECL void set_par(int idx, const d_timestamp& val);
   DLL_DECL void set_par(int idx, const d_binary& val);
   DLL_DECL void set_par(int idx, const d_string& val);
   DLL_DECL void set_par(int idx); // null

   DLL_DECL void execute();
   DLL_DECL bool fetch();
   DLL_DECL bool close();

   int get_num_cols() const
      { return columns.size() - 1; }

   SQLSMALLINT get_col_sql_type(int idx) const
      { return columns[idx].sql_type; }

   const SQLWCHAR* get_col_name(int idx) const
      { return columns[idx].name; }

   SQLSMALLINT get_col_name_len(int idx) const
      { return columns[idx].name_len; }

   int get_num_pars() const
      { return pars.size(); }

   SQLSMALLINT get_par_sql_type(int idx) const
      { return pars[idx-1].sql_type; }
   SQLUINTEGER get_par_col_size(int idx) const
      { return pars[idx-1].col_size; }
   SQLSMALLINT get_par_num_dec_digits(int idx) const
      { return pars[idx-1].dec_digits; }
   SQLSMALLINT is_par_nullable(int idx) const
      { return pars[idx-1].nullable; }
   bool is_par_unbound(int idx) const
      { return pars[idx-1].unbound; }
};

//==============

const int res_set_buf_size = 32 * 1024;

class Res_set_reader {
private:
   static char* buf[res_set_buf_size];
   static SQLINTEGER size;

public:
   static SQLINTEGER get_size()
      { return size; }

   static bool get(SQLHSTMT hstmt, SQLUSMALLINT idx, SQLSMALLINT cpp_type);

   static void put(d_binary* val)
      {
         byte_size_t _size = (byte_size_t) size; // VC7 conversion warning
         val->append_bin((char *) buf,
                         (_size < res_set_buf_size) ? _size : res_set_buf_size);
      }
   static void put(d_wstring* val)
      {
         byte_size_t _size = (byte_size_t) size; // VC7 conversion warning
         char_size_t len =
            ((_size < res_set_buf_size) ? _size : res_set_buf_size) / sizeof(wchar_t);
         val->append((wchar_t *) buf, len);
      }
};

//==============

class d_seq_query : public abs_d_query {
private:
   int idx;

   template<typename From, typename To>
   void get_fixed_size_data(int idx, From* buf, To* val);

   template<typename T>
   void get_var_size_data_low(int idx, T* val, SQLSMALLINT cpp_type);

   void get_var_size_data(int idx, d_binary* val)
      {
         val->assign("", 0);
         get_var_size_data_low(idx, val, SQL_C_BINARY);
      }
   void get_var_size_data(int idx, d_string* val, str_conv_t conv);
   void get_var_size_data(int idx, d_wstring* val);

   void __get_data(int idx, d_int* val)
      {
         int buf;
         get_fixed_size_data(idx, &buf, val);
      }
   void __get_data(int idx, d_double* val)
      {
         double buf;
         get_fixed_size_data(idx, &buf, val);
      }
   void __get_data(int idx, d_bool* val)
      {
         int buf;
         get_fixed_size_data(idx, &buf, val);
      }
   void __get_data(int idx, d_currency* val)
      {
         double buf;
         get_fixed_size_data(idx, &buf, val);
      }
   void __get_data(int idx, d_time* val)
      {
         TIME_STRUCT buf;
         get_fixed_size_data(idx, &buf, val);
      }
   void __get_data(int idx, d_date* val)
      {
         DATE_STRUCT buf;
         get_fixed_size_data(idx, &buf, val);
      }
   void __get_data(int idx, d_timestamp* val)
      {
         TIMESTAMP_STRUCT buf;
         get_fixed_size_data(idx, &buf, val);
      }
   void __get_data(int idx, d_binary* val)
      { get_var_size_data(idx, val); }
   void __get_data(int idx, d_string* val, str_conv_t conv)
      { get_var_size_data(idx, val, conv); }
   void __get_data(int idx, d_wstring* val)
      { get_var_size_data(idx, val); }

public:
   d_seq_query()
      : idx(1)
      {}

   d_seq_query(Database* db)
      : abs_d_query(db), idx(1)
      {}

   bool fetch()
      { idx = 1; return abs_d_query::fetch(); }

   DLL_DECL void get_data(d_int* val);
   DLL_DECL void get_data(d_double* val);
   DLL_DECL void get_data(d_bool* val);
   DLL_DECL void get_data(d_currency* val);
   DLL_DECL void get_data(d_time* val);
   DLL_DECL void get_data(d_date* val);
   DLL_DECL void get_data(d_timestamp* val);
   DLL_DECL void get_data(d_binary* val);
   DLL_DECL void get_data(d_string* val, str_conv_t conv = NO_CONV);
   DLL_DECL void get_data(d_wstring* val);

   void skip(unsigned int num_cols = 1)
      { idx += num_cols; }
   int get_cur_idx() const
      { return idx; }
};

template<typename From, typename To>
void d_seq_query::get_fixed_size_data(int idx, From* buf, To* val)
{
   SQLINTEGER size;

   int rc = SQLGetData(hstmt,
                       idx,
                       val->get_cpp_type(),
                       buf,
                       sizeof(From),
                       &size);
   if (rc != SQL_SUCCESS) throw_err(SQL_HANDLE_STMT, hstmt);

   if (size == SQL_NULL_DATA)
      val->make_null();
   else
      *val = *buf;
}

//==============

typedef d_seq_query d_query;

#endif
