#ifndef CPPBIND_T_IOSTREAM_H
#define CPPBIND_T_IOSTREAM_H

#include "t_fmt.h"
#include "t_streambuf.h"
#include "uni_conv.h"

//==============

typedef std::istream istream;
typedef std::ostream ostream;

class t_istream : protected istream {
private:
   dlist_in_fmt dlist_fmt;
   str_in_fmt str_fmt;

   t_ibuf* buf;
   in_fmt* fmt;

   std::streambuf* rdbuf(std::streambuf*);

   t_istream(const t_istream&);
   t_istream& operator=(const t_istream&);

public:
   t_istream(t_ibuf* ibuf)
      : istream(ibuf), buf(ibuf), fmt(&dlist_fmt)
      {}
   t_istream()
      : istream(0), buf(0), fmt(&dlist_fmt)
      {}

   void set_buf(t_ibuf* ibuf)
      { clear(); istream::rdbuf(ibuf); buf = ibuf; }

   void set_fmt_dlist()
      { fmt = &dlist_fmt; }
   void set_fmt_str(bool uni)
      { str_fmt.set_uni(uni); fmt = &str_fmt; }

   virtual bool at_end()
      { return buf->showmanyc() == 0; }

   virtual void ignore_rest()
      { ignore(buf->showmanyc()); }

   void skip_elem()
      { fmt->skip_elem(*this); }

   void get(__int64* pval, bool* null_val)
      { fmt->get(*this, pval, null_val); }
   void get(double* pval, bool* null_val)
      { fmt->get(*this, pval, null_val); }

   void get(bool* is_uni, char* buf, byte_size_t cap, byte_size_t* psize)
      { fmt->get(*this, is_uni, buf, cap, psize); }
   void get(bool* is_uni, std::string* str)
      { fmt->get(*this, is_uni, str); }
   void get(std::string* str)
      { fmt->get(*this, str); }
};

#ifndef DECUNIX
inline t_istream& operator>>(t_istream& in, __int64& val)
{ bool null; in.get(&val, &null); return in; }
#endif

inline t_istream& operator>>(t_istream& in, int& val)
{ __int64 v; bool null; in.get(&v, &null); val = int(v); return in; }

inline t_istream& operator>>(t_istream& in, long& val)
{ __int64 v; bool null; in.get(&v, &null); val = long(v); return in; }

inline t_istream& operator>>(t_istream& in, bool& val)
{ __int64 v; bool null; in.get(&v, &null); val = (v == 0) ? false : true; return in; }

inline t_istream& operator>>(t_istream& in, double& val)
{ bool null; in.get(&val, &null); return in; }

inline t_istream& operator>>(t_istream& in, std::string& val)
{ in.get(&val); return in; }

//==============

class t_ostream : protected ostream, public is_uni_srv_info {
private:
   out_fmt* fmt;

   dlist_out_fmt dlist_fmt;

   void set_fmt()
      { dlist_fmt.set_output(this); fmt = &dlist_fmt; }

   t_ostream& operator=(const t_ostream&);
   t_ostream(const t_ostream&);

public:
   t_ostream(std::streambuf* buf)
      : ostream(buf)
      { set_fmt(); }
   t_ostream()
      : ostream(0)
      { set_fmt(); }

   void put(__int64 val)
      { fmt->put(val); }
   void put(double val)
      { fmt->put(val); }

   void put(uni_t uni, const char* buf, byte_size_t len)
      { fmt->put(uni, buf, len); }

   void put_null()
      { fmt->put_null(); }

   // std::ostream inteface

   t_ostream& write(const char* buf, std::streamsize len)
      { ostream::write(buf, len); return *this; }
};

#ifndef DECUNIX
inline t_ostream& operator<<(t_ostream& out, __int64 val)
{ out.put(val); return out; }
#endif

inline t_ostream& operator<<(t_ostream& out, int val)
{ out.put(__int64(val)); return out; }

inline t_ostream& operator<<(t_ostream& out, size_t val)
{ out.put(__int64(val)); return out; }

inline t_ostream& operator<<(t_ostream& out, bool val)
{ out.put(__int64(val)); return out; }

inline t_ostream& operator<<(t_ostream& out, long val)
{ out.put(__int64(val)); return out; }

inline t_ostream& operator<<(t_ostream& out, double val)
{ out.put(val); return out; }

inline t_ostream& operator<<(t_ostream& out, const_name_t val)
{
   if (out.is_uni_srv())
      out.put(cpp_uni, (char*) val, wcslen(val) * sizeof(wchar_t));
   else
   {
      int size = UnicodeToMultiByte(val, wcslen(val), conv_buf, conv_buf_size);
      out.put(multibyte, conv_buf, size);
   }
   return out;
}

//==============

class t_istrstream : public t_istream {
private:
   t_istrbuf ibuf;

   void set_buf(t_ibuf* ibuf)
      { t_istream::set_buf(ibuf); }

public:
   t_istrstream(const char* buf, std::streamsize size)
      : ibuf(buf, size)
      { set_buf(&ibuf); }
   t_istrstream()
      { set_buf(&ibuf); }

   void set_buf(const char* buf, std::streamsize size)
      { clear(); ibuf.set_buf(buf, size); }

   // string interface

   const char* get_buf() const
      { return ibuf.get_buf(); }
   std::streamsize get_size() const
      { return ibuf.get_size(); }

   void rewind()
      { clear(); ibuf.rewind(); }
};

//==============

class t_ostrstream : public t_ostream {
private:
   t_ostrbuf obuf;

public:
   t_ostrstream()
      { rdbuf(&obuf); }

   // degenerate case, don't copy the buffer
   t_ostrstream(const t_ostrstream& out)
      { rdbuf(&obuf); }
   t_ostrstream& operator=(const t_ostrstream& out)
      { return *this; }

   const char* get_buf()
      { return obuf.get_buf(); }
   std::streamsize get_size()
      { return obuf.get_size(); }

   void reset()
      { clear(); obuf.reset(); }
};

//==============

class t_itcpstream : public t_istream {
private:
   t_itcpbuf ibuf;

   void set_buf(t_ibuf* ibuf)
      { t_istream::set_buf(ibuf); }

public:
   t_itcpstream()
      { set_buf(&ibuf); }

   void set_messenger(void* pm)
      { ibuf.set_messenger(pm); }

   void reset()
      { ibuf.reset(); }
   void get_header()
      { ibuf.get_header(); }

   bool at_end()
      { return ibuf.at_end(); }
   void ignore_rest()
      { ibuf.ignore_rest(); }
};

//==============

class t_otcpstream : public t_ostream {
private:
   t_otcpbuf obuf;

public:
   t_otcpstream()
      { rdbuf(&obuf); }

   void set_messenger(void* pm)
      { obuf.set_messenger(pm); }

   // degenerate case, don't copy the buffer
   t_otcpstream(const t_otcpstream& out)
      { rdbuf(&obuf); }
   t_otcpstream& operator=(const t_otcpstream& out)
      { return *this; }

   void reset()
      { clear(); obuf.reset(); }
   void send(const char* msg)
      { obuf.send(msg); }
};

inline t_otcpstream& operator<<(t_otcpstream& out, const_name_t val)
{ static_cast<t_ostream&>(out) << val; return out; }

#endif
